# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import dash_bootstrap_components as dbc
import uuid

from dash_extensions.enrich import html


class ProjectInformationTableAIO(html.Div):

    class ids:
        pass

    ids = ids

    def __init__(self, data: dict, aio_id: str = None):
        """ProjectInformationTableAIO is an All-in-One component that is composed
        of multiple components which are not exposed but only configured internally.

        - `data` - Data to be displayed in the table and pie chart components.
        - `aio_id` - The All-in-One component ID used to generate the table components's dictionary IDs.
        """

        if aio_id is None:
            aio_id = str(uuid.uuid4())

        self._style: dict = {
            "textAlign": "left",
            "fontSize": 12,
            "fontFamily": "Roboto",
            "color": "rgba(0, 0, 0, 1)",
            "backgroundColor": "rgba(255, 255, 255, 0)"
        }

        table_data = self.to_dash(data)

        super().__init__([
            html.H5("Project information", className="card-title"),
            html.Hr(className="my-2"),
            dbc.Card(
                [
                    dbc.CardBody(
                        table_data,
                        className="border-0 bg-transparent"
                    ),
                ],
                className="border-0 bg-transparent"
            )
        ])

    def to_dash(self, data: dict) -> list:

        table_data = []

        for key, value in data.items():
            table_data.append(
                dbc.Row(
                    [
                        dbc.Col(
                            html.Label(
                                key,
                                style = self._style
                            ),
                            width=1
                        ),
                        dbc.Col(
                            html.Label(
                                value,
                                style = self._style
                            ),
                            width=11
                        ),
                    ]
                )
            )

        return table_data



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the visualization view."""

import json
import optislang_dash_lib

from dash_extensions.enrich import html

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep


def layout(problem_setup_step: ProblemSetupStep, monitoring_step: MonitoringStep) -> html.Div:
    """Layout of the visualization view."""

    full_project_status_info = json.loads(problem_setup_step.full_project_status_info_file.read_text())

    return html.Div(
        [
            optislang_dash_lib.VisualizationComponent(
                id="visualization-component",
                project_state=full_project_status_info,
                system_id=monitoring_step.selected_actor_from_treeview,
                state_idx=0,
            ),
        ]
    )



{%- if cookiecutter.copyright != "None" -%}
# Copyright (c) {% now "utc", '%Y' %}, {{ cookiecutter.copyright }}. Unauthorised use, distribution or duplication is prohibited
{% endif %}

import json
import multiprocessing
import random

import requests

from src.server import create_app


class TestServer:
    def setup_class(self):
        port = random.randint(5000, 10000)
        self.base_url = "http://127.0.0.1:" + str(port)
        app = create_app()
        self.server_proc = multiprocessing.Process(target=app.run, args=["127.0.0.1", port, False])
        self.server_proc.start()

    def teardown_class(self):
        self.server_proc.terminate()
        self.server_proc.join()

    def test_get_version(self):
        response = requests.get(self.base_url + "/api/version", timeout=3)
        assert response.status_code == 200

    def test_get_health_status(self):
        response = requests.get(self.base_url + "/api/health", timeout=3)
        assert response.status_code == 200
        assert json.loads(response.content)["status"] == "ok"



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import dash_bootstrap_components as dbc
import pandas as pd
import uuid

from dash_extensions.enrich import html, dash_table

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.utilities.common_functions import sorted_nicely


class DesignTableAIO(html.Div):

    class ids:
        datatable = lambda aio_id: {
            'component': 'DesignTableAIO',
            'subcomponent': 'datatable',
            'aio_id': aio_id
        }

    ids = ids

    def __init__(self, data: dict, aio_id: str = None):
        """DesignTableAIO is an All-in-One component that is composed
        of a `dash_table.DataTable` component.

        - `data` - Data to be displayed in the datatable component.
        - `aio_id` - The All-in-One component ID used to generate the table components's dictionary IDs.
        """

        if aio_id is None:
            aio_id = str(uuid.uuid4())

        data = self.transform_data(data)

        datatable_props = {
            "data": data.to_dict('records'),
            "columns": [{"name": i, "id": i, "type": "text"} for i in data.columns],
            "fixed_rows": {"headers": True},
            "style_header": {
                "textAlign": "center",
                "font_family": "Roboto",
                "font_size": "15px",
                "fontWeight": "bold",
            },
            "style_cell": {
                "textAlign": "center",
                "font_family": "Roboto",
                "font_size": "15px",
                "whiteSpace": "normal",
            },
            "style_data_conditional": [
                {
                    "if": {"column_id": "Status", "filter_query": '{Status} eq "Succeeded"'},
                    "backgroundColor": "rgb(223, 240, 208)",
                    "color": "rgb(0, 0, 0)",
                },
                {
                    "if": {"column_id": "Status", "filter_query": '{Status} eq "Not succeeded"'},
                    "backgroundColor": "rgb(254, 221, 215)",
                    "color": "rgb(0, 0, 0)",
                },
                {
                    "if": {"column_id": "Feasible", "filter_query": '{Feasible} eq "True"'},
                    "backgroundColor": "rgb(223, 240, 208)",
                    "color": "rgb(0, 0, 0)",
                },
                {
                    "if": {"column_id": "Feasible", "filter_query": '{Feasible} eq "False"'},
                    "backgroundColor": "rgb(254, 221, 215)",
                    "color": "rgb(0, 0, 0)",
                },
            ],
            "style_as_list_view": True,
        }

        super().__init__([
            html.Br(),
            dbc.Row(
                [
                    dbc.Col(
                        dash_table.DataTable(id=self.ids.datatable(aio_id), **datatable_props),
                        width=12
                    ),
                ]
            )
        ])

    def sort_designs(self, designs: list) -> list:

        sorted_designs = sorted_nicely(designs)
        return [designs.index(design) for design in sorted_designs]


    def transform_data(self, data: dict) -> pd.DataFrame:

        data = pd.DataFrame(data)

        if len(data["Design"]):

            # OptiSLang returns the designs in an unsorted way.
            # The following method organizes the rows in ascending order.
            data = data.loc[self.sort_designs(data["Design"].to_list())]

            for col, dtype in data.dtypes.items():
                if dtype == "bool":
                    data[col] = data[col].astype("str")

            data = data.round(6)

        return data



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import dash_bootstrap_components as dbc
import uuid

from dash_extensions.enrich import html, dcc


class ActorInformationTableAIO(html.Div):

    class ids:
        pie_chart = lambda aio_id: {
            'component': 'ActorStatisticsTableAIO',
            'subcomponent': 'pie_chart',
            'aio_id': aio_id
        }

    ids = ids

    def __init__(self, data: dict, aio_id: str = None):
        """ActorInformationTableAIO is an All-in-One component that is composed
        of a `px.pie` component (plotly express pie chart) and additional components
        which are not exposed but only configured internally.

        - `data` - Data to be displayed in the table and pie chart components.
        - `aio_id` - The All-in-One component ID used to generate the table components's dictionary IDs.
        """

        if aio_id is None:
            aio_id = str(uuid.uuid4())

        self._style: dict = {
            "textAlign": "left",
            "fontSize": 12,
            "fontFamily": "Roboto",
            "color": "rgba(0, 0, 0, 1)",
            "backgroundColor": "rgba(255, 255, 255, 0)"
        }

        table_data, pie_chart_data = self.to_dash(data)

        if pie_chart_data:
            pie_chart_props = {
                "figure": {
                    "data": [
                        {
                            "type": "pie",
                            "labels": pie_chart_data["Category"],
                            "values": pie_chart_data["Count"],
                            "hole": 0.8,
                            "textinfo": None,
                            "showlegend": True,
                            "marker": {
                                "colors": [
                                    'rgba(173, 191, 11, 1)',
                                    'rgba(248, 105, 80, 1)',
                                    'rgba(251, 201, 93, 1)'
                                ]
                            }
                        }
                    ],
                    "layout": {
                        "autosize": False,
                        "width": 270,
                        "height": 270,
                        "plot_bgcolor": 'rgba(0,0,0,0)',
                        "paper_bgcolor": 'rgba(0,0,0,0)',
                        "margin": {
                            "l": 0,
                            "r": 0,
                            "b": 0,
                            "t": 0,
                        },
                        "legend": {
                            "orientation": "h"
                        }
                    }
                }

            }

        super().__init__([
            html.H5("Actor information", className="card-title"),
            html.Hr(className="my-2"),
            dbc.Row(
                [
                    dbc.Col(
                        dbc.Card(
                            [
                                dbc.CardBody(
                                    table_data,
                                    className="border-0 bg-transparent"
                                ),
                            ],
                            className="border-0 bg-transparent"
                        ),
                        width=7
                    ),
                    dbc.Col(
                        dbc.Card(
                            [
                                dbc.CardBody(
                                    dcc.Graph(id=self.ids.pie_chart(aio_id), **pie_chart_props) if pie_chart_data else None,
                                    className="border-0 bg-transparent"
                                ),
                            ],
                            className="border-0 bg-transparent"
                        ),
                        width=5
                    ),

                ]
            )
        ])

    def to_dash(self, data: dict) -> list:

        # Collect table data
        table_data = []
        for key, value in data.items():
            table_data.append(
                dbc.Row(
                    [
                        dbc.Col(
                            html.Label(
                                key,
                                style = self._style
                            ),
                            width=2
                        ),
                        dbc.Col(
                            html.Label(
                                value,
                                style = self._style
                            ),
                            width=10
                        ),
                    ]
                )
            )

        # Collect pie chart data
        pie_chart_data = {}
        if "Processed" in data.keys():
            pie_chart_data = {'Category': ['Success', 'Failure', 'Pending'], 'Count': [0, 0, 0]}
            for index, state in enumerate(["Succeeded", "Not succeeded", "Pending"]):
                pie_chart_data["Count"][index] = data[state]

        return table_data, pie_chart_data



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the second step."""


from dash_extensions.enrich import dcc, html

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.second_step import SecondStep


def layout(step: SecondStep):
    """Layout of the second step UI."""
    return html.Div(
        [
            dcc.Markdown("""#### Second step""", className="display-3"),
            dcc.Markdown("""###### Subtitle.""", className="display-3"),
            html.Hr(className="my-2"),
            html.Br(),
        ]
    )



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Application."""

import dash_bootstrap_components as dbc
import dash_uploader as du
import os
import tempfile

from dash_extensions.enrich import DashProxy, MultiplexerTransform, NoOutputTransform, TriggerTransform


class Dash(DashProxy):
    """Custom Dash class."""
    def interpolate_index(self, **kwargs):

        # Get the original index content (including default Dash CSS styles) from the parent class
        index_str = super().interpolate_index(**kwargs)

        # Combine the original index content and your custom CSS
        modified_index = index_str.replace('</html>', f'<html class="light">')

        return modified_index

app = Dash(
    __name__,
    external_stylesheets=[dbc.themes.BOOTSTRAP, dbc.icons.FONT_AWESOME],
    suppress_callback_exceptions=True,
    transforms=[NoOutputTransform(), TriggerTransform(), MultiplexerTransform()],
)

# If folder doesn't exist, it will be created later
UPLOAD_DIRECTORY = os.path.join(tempfile.gettempdir(), "GLOW")
du.configure_upload(app, UPLOAD_DIRECTORY)

# !IMPORTANT Keeping the import line here to adapt with dash_uploader config, moving the import above will fail the
# dash uploader configuration
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.pages.page import layout

app.layout = layout



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the project summary step."""

import dash_bootstrap_components as dbc
import json
from json.decoder import JSONDecodeError
import logging

from dash.exceptions import PreventUpdate
from dash_extensions.enrich import html, Input, Output, State, dcc
from ansys.saf.glow.client.dashclient import DashClient, callback
from ansys.saf.glow.core.method_status import MethodStatus

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.node_control import NodeControlAIO
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.project_information_table import ProjectInformationTableAIO
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.button_group import ButtonGroup
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.utilities.common_functions import LOG_MESSAGE_COLORS


def layout(problem_setup_step: ProblemSetupStep, monitoring_step: MonitoringStep) -> html.Div:
    """Layout of the project summary view."""
    alert_props = {
        "children":monitoring_step.project_command_execution_status["alert-message"],
        "color":monitoring_step.project_command_execution_status["alert-color"],
        "is_open": bool(monitoring_step.project_command_execution_status["alert-message"]),
    }
    button_group = ButtonGroup(options=monitoring_step.project_btn_group_options, disabled=monitoring_step.commands_locked).buttons

    # Get project data
    project_data = json.loads(problem_setup_step.project_data_file.read_text())

    return html.Div(
        [
            html.Br(),
            html.Div(
                id="alerts_container",
                style={"position": "fixed", "top": 130, "right": 10, "width": 350},
            ),
            dbc.Row(
                [
                    dbc.Col(
                        html.Div(
                            ProjectInformationTableAIO(
                                project_data["project"]["information"],
                            ),
                            id="project_information_table"
                        ),
                        width=12
                    )
                ]
            ),
            html.Br(),
            dbc.Row(
                [
                    dbc.Col(NodeControlAIO(button_group, alert_props, "project-commands"), width=12),
                ]
            ),
            dcc.Interval(
                id="project_summary_auto_update",
                interval=monitoring_step.auto_update_frequency,  # in milliseconds
                n_intervals=0,
                disabled=False if monitoring_step.auto_update_activated else True
            ),
        ]
    )


@callback(
    Output("alerts_container", "children"),
    Input("project_summary_auto_update", "n_intervals"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def display_alerts(n_intervals, pathname):
    # Get project
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    # Get problem setup step
    problem_setup_step = project.steps.problem_setup_step

    alerts_container = []

    method_state = problem_setup_step.get_long_running_method_state("start_and_monitor_osl_project")
    if method_state.status == MethodStatus.Failed:
        alerts_container.append(
            dbc.Toast(
                method_state.exception_message,
                header="Error",
                is_open=True,
                dismissable=True,
                icon="danger",
            )
        )

    if len(problem_setup_step.alerts) > 0:
        if problem_setup_step.alerts:
            alerts_container.append(
                dbc.Toast(
                    problem_setup_step.alerts["message"],
                    header=problem_setup_step.alerts["level"].capitalize(),
                    is_open=True,
                    dismissable=True,
                    icon=LOG_MESSAGE_COLORS[problem_setup_step.alerts["level"]],
                )
            )
        return alerts_container
    else:
        raise PreventUpdate


@callback(
    Output("project_information_table", "children"),
    Output("selected_state_dropdown", "options"),
    Output("selected_state_dropdown", "value"),
    Output("project_summary_auto_update", "disabled"),
    Input("project_summary_auto_update", "n_intervals"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def update_view(n_intervals, pathname):
    """Update design table."""
    # Get project
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    # Get problem setup step
    problem_setup_step = project.steps.problem_setup_step
    # Get monitoring step
    monitoring_step = project.steps.monitoring_step

    if monitoring_step.auto_update_activated:
        try:
            # Get project data
            project_data = json.loads(problem_setup_step.project_data_file.read_text())
            # Collect states ids
            if not monitoring_step.selected_state_id:
                if len(project_data.get("actors").get(monitoring_step.selected_actor_from_treeview).get("states_ids")):
                    monitoring_step.selected_state_id = project_data.get("actors").get(monitoring_step.selected_actor_from_treeview).get("states_ids")[0]
            return (
                ProjectInformationTableAIO(project_data.get("project").get("information")),
                project_data.get("actors").get(monitoring_step.selected_actor_from_treeview).get("states_ids"),
                monitoring_step.selected_state_id,
                True if problem_setup_step.osl_project_state in ["NOT STARTED", "FINISHED", "ABORTED"] else False
            )
        except Exception as e:
            logging.error(str(e))
            raise PreventUpdate
    else:
        raise PreventUpdate



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the node status overview view."""

import json
import optislang_dash_lib

from dash_extensions.enrich import html

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.problem_setup_step import ProblemSetupStep


def layout(problem_setup_step: ProblemSetupStep) -> html.Div:
    """Layout of the status overview view."""

    full_project_status_info = json.loads(problem_setup_step.full_project_status_info_file.read_text())

    if full_project_status_info:
        return html.Div(
            id="node-status-view",
            children=[
                optislang_dash_lib.Nodestatusviewcomponent(
                    id="node-status-view-component",
                    project_state=full_project_status_info,
                ),
            ],
        )
    else:
        return html.Div([])



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import dash_bootstrap_components as dbc
import pandas as pd
import uuid

from dash_extensions.enrich import html, dash_table


class ActorLogsTableAIO(html.Div):

    class ids:
        datatable = lambda aio_id: {
            'component': 'ActorLogsTableAIO',
            'subcomponent': 'datatable',
            'aio_id': aio_id
        }

    ids = ids

    def __init__(self, data: dict, store_value: int= None, aio_id: str = None):
        """ActorLogsTableAIO is an All-in-One component that is composed
        of a `dash_table.DataTable` component.

        - `data` - Data to be displayed in the datatable component.
        - `aio_id` - The All-in-One component ID used to generate the table components's dictionary IDs.
        """

        if aio_id is None:
            aio_id = str(uuid.uuid4())

        data = pd.DataFrame(data)

        datatable_props = {
            "data": data.to_dict('records'),
            "columns": [{"name": i, "id": i, "type": "text"} for i in data.columns],
            "fixed_rows": {"headers": True},
            "style_header": {"font_family": "Roboto", "font_size": "15px", "fontWeight": "bold"},
            "style_cell": {
                "textAlign": "left",
                "font_family": "Roboto",
                "font_size": "15px",
            },
            "style_as_list_view": True,
            "page_current": store_value
        }

        super().__init__([
            html.H5("Actor Log", className="card-title"),
            html.Hr(className="my-2"),
            dbc.Card(
                [
                    dbc.CardBody(
                        [
                            dash_table.DataTable(id=self.ids.datatable(aio_id), **datatable_props),
                        ],
                        className="border-0 bg-transparent"
                    ),
                ],
                className="border-0 bg-transparent"
            )
        ])



{%- if cookiecutter.copyright != "None" -%}
# Copyright (c) {% now "utc", '%Y' %}, {{ cookiecutter.copyright }}. Unauthorised use, distribution or duplication is prohibited
{% endif %}

"""
{{ cookiecutter.project_name }}.

{{ cookiecutter.library_name }}
"""
from __future__ import print_function

import logging
import os
from pathlib import Path
import sys

import grpc

sys.path.insert(0, os.path.join(Path(os.path.dirname(os.path.realpath(__file__))).parent, "stubs"))

import stubs.pingserver_pb2 as pb2
import stubs.pingserver_pb2_grpc as pb2_grpc


def run():
    """Run client."""
    with grpc.insecure_channel("localhost:50051") as channel:
        stub = pb2_grpc.PingerStub(channel)
        response = stub.WhoPing(pb2.UserRequest(name="you"))
    print(response.message)


if __name__ == "__main__":
    logging.basicConfig()
    run()



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Backend of the problem setup step."""

import json
from pathlib import Path
import time
from typing import List, Union, Optional

from ansys.optislang.core import logging
from ansys.optislang.core.errors import OslCommunicationError
from ansys.optislang.core.osl_server import OslServer
from ansys.saf.glow.solution import FileReference, StepModel, StepSpec, instance, long_running, transaction

from ansys.solutions.{{cookiecutter.__solution_name_slug}}.datamodel import datamodel
from ansys.solutions.{{cookiecutter.__solution_name_slug}}.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.{{cookiecutter.__solution_name_slug}}.solution.optislang_manager import OptislangManager
from ansys.solutions.{{cookiecutter.__solution_name_slug}}.utilities.common_functions import read_log_file, check_optislang_server, get_states_ids_from_states


class MonitoringStep(StepModel):
    """Step model of the monitoring step."""

    # Parameters ------------------------------------------------------------------------------------------------------

    # Frontend persistence
    selected_actor_from_treeview: Optional[str] = None # uid of the actor selected from the treeview.
    selected_command: Optional[str] = None
    selected_actor_from_command: Optional[str] = None # uid of the actor from which a control command has been requested.
    selected_state_id: Optional[str] = None
    commands_locked: bool = False
    auto_update_activated: bool = True
    actor_uid: Optional[str] = None
    selected_page: int = 0
    project_command_execution_status: dict = {"alert-message": "", "alert-color": "info"}
    actor_command_execution_status: dict = {"alert-message": "", "alert-color": "info"}
    project_btn_group_options: List = [
                                {
                                    "icon": "fas fa-play",
                                    "tooltip": "Restart optiSLang project.",
                                    "value": "restart",
                                    "id": {
                                        "type":"action-button",
                                        "action":"restart"
                                    }
                                },
                                {
                                    "icon":"fa fa-hand-paper",
                                    "tooltip": "Stop optiSLang project gently.",
                                    "value": "stop_gently",
                                    "id": {
                                        "type":"action-button",
                                        "action":"stop_gently"
                                    }
                                },
                                {
                                    "icon": "fas fa-stop",
                                    "tooltip":"Stop optiSLang project.",
                                    "value": "stop",
                                    "id": {
                                        "type": "action-button",
                                        "action":"stop"
                                        }
                                    },
                                {
                                    "icon": "fas fa-fast-backward",
                                    "tooltip": "Reset optiSLang project.",
                                    "value": "reset",
                                    "id": {
                                        "type":"action-button",
                                        "action":"reset"
                                        }
                                    },
                                {
                                    "icon": "fas fa-power-off",
                                    "tooltip": "Shutdown optiSLang project.",
                                    "value": "shutdown",
                                    "id": {
                                        "type":"action-button",
                                        "action":"shutdown"
                                    }
                                },
                            ]
    actor_btn_group_options: List = [
                            {
                                "icon": "fas fa-play",
                                "tooltip": "Restart node.",
                                "value": "restart",
                                "id": {
                                    "type":"action-button",
                                    "action":"restart"
                                }
                            },

                            {
                                "icon":"fa fa-hand-paper",
                                "tooltip": "Stop node gently.",
                                "value": "stop_gently",
                                "id": {
                                    "type":"action-button",
                                    "action":"stop_gently"
                                }
                            },

                            {
                                "icon": "fas fa-stop",
                                "tooltip":"Stop node.",
                                "value": "stop",
                                "id": {
                                    "type": "action-button",
                                    "action":"stop"
                                }
                            },

                            {
                                "icon": "fas fa-fast-backward",
                                "tooltip": "Reset node.",
                                "value": "reset",
                                "id": {
                                    "type":"action-button",
                                    "action":"reset"
                                }
                            },
                        ]

    # Backend data model
    auto_update_frequency: float = 2000 # millisecond
    command_timeout: int = 30 # second

    # File storage ----------------------------------------------------------------------------------------------------

    # Methods ---------------------------------------------------------------------------------------------------------

    @transaction(
        self=StepSpec(
            download=[
                "command_timeout",
                "selected_actor_from_command",
                "selected_command",
            ],
            upload=["actor_uid"],
        )
    )
    @instance("problem_setup_step.osl_manager", identifier="osl_manager")
    def control_node_state(self, osl_manager: OptislangManager) -> None:
        """Update the state of root or actor node based on the selected command in the UI."""
        osl = osl_manager.instance
        if not self.selected_command == "shutdown":
            if self.selected_actor_from_command == osl.project.root_system.uid:
                node = osl.project.root_system
                self.actor_uid = None
            else:
                node = osl.project.root_system.find_node_by_uid(self.selected_actor_from_command, search_depth=-1)
                self.actor_uid = node

            status = node.control(self.selected_command, wait_for_completion=True, timeout=self.command_timeout)

            if not status:
                raise Exception(f"{self.selected_command.replace('_', ' ').title()} command against node {node.get_name()} failed.")
        else:
            osl_manager.shutdown()




# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

from ansys.optislang.core import Optislang, utils, logging
from pathlib import Path
from fastapi import Body, FastAPI


OSL_SERVER_PORT = None


app = FastAPI()


@app.post("/start")
async def start_instance(project_path: str = Body(...), project_properties_file: str = Body(...), osl_version: int = Body(...), loglevel: str = Body(...)):
    global OSL_SERVER_PORT
    osl = Optislang(
        project_path=project_path,
        executable=utils.get_osl_exec(osl_version)[1],
        loglevel=loglevel,
        reset=True,
        auto_relocate=True,
        shutdown_on_finished=False,
        import_project_properties_file=project_properties_file,
        ini_timeout=300
    )
    # Configure logging.
    osl_logger = logging.OslLogger(
        loglevel=loglevel,
        log_to_file=True,
        logfile_name=Path(project_path).parent / "optiSLang.log",
        log_to_stdout=True,
    )
    osl.__logger = osl_logger.add_instance_logger(osl.name, osl, loglevel)
    osl.log.info("Start instance")
    # Get server host
    osl_server_host = osl.get_osl_server().get_host()
    # Get server port
    server_info = osl.get_osl_server().get_server_info()
    osl_server_port = server_info["server"]["server_port"]
    OSL_SERVER_PORT = osl_server_port
    return {"host": osl_server_host, "port": osl_server_port}


@app.get("/")
async def get_port():
    return {"port": OSL_SERVER_PORT}


@app.get("/health")
async def health():
    return "healthy"



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

from typing import Optional, Union

from ansys.optislang.core import Optislang
from ansys.saf.glow.core.instance_manager import AbstractInstanceIdentificationClient
from ansys.saf.glow.solution import FileReference, InstanceManager
from ansys.saf.glow.storage.base import AbstractStoragePath
import httpx

PRODUCT_NAME = "optislang"
SERVICE_NAME = "http"


class OptislangManager(InstanceManager[Optislang]):
    def __init__(
        self,
        instance_identification: AbstractInstanceIdentificationClient,
        method_directory: Optional[AbstractStoragePath] = None,
    ):
        super().__init__(instance_identification, method_directory)

    def initialize(
        self,
        version: Optional[str] = None,
        project_path: Optional[Union[FileReference, str]] = None,
        project_properties_file: Optional[Union[FileReference, str]] = None,
        osl_version: Optional[int] = None,
        loglevel: Optional[str] = None,
    ):
        """Initialize and start the optiSLang server."""
        self.initialize_service(SERVICE_NAME, version)

        response = httpx.post(
            f"{self._service.uri}/start",
            json={
                "project_path": project_path,
                "project_properties_file": project_properties_file,
                "osl_version": osl_version,
                "loglevel": loglevel
            },
            timeout=300
        ).json()

    @classmethod
    def get_product_name_implement(cls) -> str:
        """Return the name of the product in the PIM environment."""
        return PRODUCT_NAME

    def implement_get_client_object(self, hostname: str, port: int) -> Optislang:
        """Return an instance of the custom product's client given the host and port of the shared product instance."""
        response = httpx.get(f"http://{hostname}:{port}").json()
        return Optislang(host=hostname, port=response["port"])

    def _get_project_file_path(self, protected_state_directory_path: str) -> str:
        pass

    def save_state_implement(self, protected_state_directory_path: str) -> None:
        """Save the state of the product instance to the given directory."""
        pass

    def load_state_implement(self, protected_state_directory_path: str) -> None:
        """Loads the state of the product instance from the given directory."""
        pass

    def shutdown_implement(self, protected_state_directory_path: str) -> None:
        """Gracefully terminates the product instance process."""
        self.instance.shutdown()
        self.instance.dispose()



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the design table view."""

import dash_bootstrap_components as dbc
import json
from json.decoder import JSONDecodeError
import logging

from dash.exceptions import PreventUpdate
from dash_extensions.enrich import html, Input, Output, State, dcc
from ansys.saf.glow.client.dashclient import DashClient, callback
from ansys.saf.glow.core.method_status import MethodStatus

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.datamodel import datamodel
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.design_table import DesignTableAIO
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.utilities.common_functions import LOG_MESSAGE_COLORS


def layout(problem_setup_step: ProblemSetupStep, monitoring_step: MonitoringStep) -> html.Div:
    """Layout of the design table view."""

    # Get project data
    project_data = json.loads(problem_setup_step.project_data_file.read_text())
    # Get actor uid
    actor_uid = monitoring_step.selected_actor_from_treeview
    # Get actor hid
    actor_hid = monitoring_step.selected_state_id
    # Collect design table data
    if monitoring_step.selected_state_id:
        design_table_data = project_data["actors"][actor_uid]["design_table"][actor_hid]
    else:
        design_table_data = datamodel.extract_design_table_data({})
    # Build layout
    return html.Div(
        [
            html.Br(),
            html.Div(
                id="alerts_container",
                style={"position": "fixed", "top": 130, "right": 10, "width": 350},
            ),
            html.Div(
                id="design_table",
                children=DesignTableAIO(
                    design_table_data,
                )
            ),
            dcc.Interval(
                id="design_table_auto_update",
                interval=monitoring_step.auto_update_frequency,  # in milliseconds
                n_intervals=0,
                disabled=False if monitoring_step.auto_update_activated else True
            ),
        ]
    )


@callback(
    Output("alerts_container", "children"),
    Input("design_table_auto_update", "n_intervals"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def display_alerts(n_intervals, pathname):
    # Get project
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    # Get problem setup step
    problem_setup_step = project.steps.problem_setup_step

    alerts_container = []

    method_state = problem_setup_step.get_long_running_method_state("start_and_monitor_osl_project")
    if method_state.status == MethodStatus.Failed:
        alerts_container.append(
            dbc.Toast(
                method_state.exception_message,
                header="Error",
                is_open=True,
                dismissable=True,
                icon="danger",
            )
        )

    if len(problem_setup_step.alerts) > 0:
        if problem_setup_step.alerts:
            alerts_container.append(
                dbc.Toast(
                    problem_setup_step.alerts["message"],
                    header=problem_setup_step.alerts["level"].capitalize(),
                    is_open=True,
                    dismissable=True,
                    icon=LOG_MESSAGE_COLORS[problem_setup_step.alerts["level"]],
                )
            )
        return alerts_container
    else:
        raise PreventUpdate


@callback(
    Output("design_table", "children"),
    Output("selected_state_dropdown", "options"),
    Output("selected_state_dropdown", "value"),
    Output("design_table_auto_update", "disabled"),
    Input("design_table_auto_update", "n_intervals"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def update_view(n_intervals, pathname):
    """Update design table."""
    # Get project
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    # Get problem setup step
    problem_setup_step = project.steps.problem_setup_step
    # Get monitoring step
    monitoring_step = project.steps.monitoring_step

    if monitoring_step.auto_update_activated:
        try:
            # Get project data
            project_data = json.loads(problem_setup_step.project_data_file.read_text())
            # Get actor uid
            actor_uid = monitoring_step.selected_actor_from_treeview
            # Get actor hid
            actor_hid = monitoring_step.selected_state_id
            # Collect design table data
            if monitoring_step.selected_state_id:
                design_table_data = project_data.get("actors").get(actor_uid).get("design_table").get(actor_hid)
            else:
                design_table_data = datamodel.extract_design_table_data({})
            # Collect states ids
            if not monitoring_step.selected_state_id:
                if len(project_data.get("actors").get(monitoring_step.selected_actor_from_treeview).get("states_ids")):
                    monitoring_step.selected_state_id = project_data.get("actors").get(monitoring_step.selected_actor_from_treeview).get("states_ids")[0]
            return (
                DesignTableAIO(design_table_data),
                project_data.get("actors").get(monitoring_step.selected_actor_from_treeview).get("states_ids"),
                monitoring_step.selected_state_id,
                True if problem_setup_step.osl_project_state in ["NOT STARTED", "FINISHED", "ABORTED"] else False
            )
        except Exception as e:
            logging.error(str(e))
            raise PreventUpdate
    else:
        raise PreventUpdate



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Common functions."""

import dash_bootstrap_components as dbc
import re

from pathlib import Path
from typing import Any, Iterable, Union, Tuple

from ansys.optislang.core.osl_server import OslServer


MONITORING_TABS = [
    {
        "label": "Project Summary",
        "tab_id": "project_summary_tab",
        "is_root": True,
        "is_system": False,
        "is_actor": False,
    },
    {"label": "Summary", "tab_id": "summary_tab", "is_root": False, "is_system": True, "is_actor": True},
    {"label": "Scenery", "tab_id": "scenery_tab", "is_root": True, "is_system": False, "is_actor": False},
    {"label": "Design Table", "tab_id": "design_table_tab", "is_root": True, "is_system": True, "is_actor": False},
    {"label": "Visualization", "tab_id": "visualization_tab", "is_root": True, "is_system": True, "is_actor": False},
    {"label": "Status Overview", "tab_id": "status_overview_tab", "is_root": True, "is_system": True, "is_actor": True},
]


PROJECT_STATES = {
    "NOT STARTED": {
        "alert": "optiSLang project not started.",
        "color": "warning"
    },
    "IDLE": {
        "alert": "optiSLang project is pending.",
        "color": "warning"
    },
    "PROCESSING": {
        "alert": "optiSLang project in progress.",
        "color": "primary"
    },
    "PAUSED": {
        "alert": "optiSLang project paused.",
        "color": "warning"
    },
    "PAUSE_REQUESTED": {
        "alert": "optiSLang project requested to pause.",
        "color": "warning"
    },
    "STOPPED": {
        "alert": "optiSLang project stopped.",
        "color": "warning"
    },
    "STOP_REQUESTED": {
        "alert": "optiSLang project requetsed to stop.",
        "color": "warning"
    },
    "GENTLY_STOPPED": {
        "alert": "optiSLang project gently stopped.",
        "color": "warning"
    },
    "GENTLE_STOP_REQUESTED": {
        "alert": "optiSLang project requested to gently stop.",
        "color": "warning"
    },
    "FINISHED": {
        "alert": "optiSLang project completed successfully.",
        "color": "success"
    },
}


LOG_MESSAGE_COLORS = {
    "info": "primary",
    "warning": "warning",
    "error": "danger"
}


def find_dicts_by_key_recursively(structure, target_level, current_path=[]):
    results = []

    if isinstance(structure, list):
        for i, item in enumerate(structure):
            path = current_path + [i]
            results.extend(find_dicts_by_key_recursively(item, target_level, path))
    elif isinstance(structure, dict):
        if "level" in structure and structure["level"] == target_level:
            results.append(current_path)
        for key, value in structure.items():
            path = current_path + [key]
            results.extend(find_dicts_by_key_recursively(value, target_level, path))

    return results


def get_dict_from_indexes_sequence(nested_structure, mixed_keys):
    """
    Retrieve a dictionary from a complex nested structure of dictionaries and lists
    using a mixed list of keys and indexes.

    Args:
        nested_structure (dict or list): The complex nested structure.
        mixed_keys (list): List of keys and indexes to follow, in mixed order.

    Returns:
        dict: The retrieved dictionary, or None if not found.
    """
    for key_or_index in mixed_keys:
        if isinstance(nested_structure, dict) and isinstance(key_or_index, str) and key_or_index in nested_structure:
            nested_structure = nested_structure[key_or_index]
        elif isinstance(nested_structure, list) and isinstance(key_or_index, int) and 0 <= key_or_index < len(nested_structure):
            nested_structure = nested_structure[key_or_index]
        else:
            return None  # Key or index not found in the current level

    if isinstance(nested_structure, dict):
        return nested_structure
    else:
        return None  # If the final element is not a dictionary


def get_treeview_items_from_project_tree(osl_project_tree: list) -> list:

    treeview_items = [
        {
            "id": "problem_setup_step",
            "text": "Problem Setup",
            "expanded": True,
            "prefixIcon": {
                "src": "https://s2.svgbox.net/hero-solid.svg?ic=adjustments"
            },
            "level": 0
        },
    ]

    for i, node in enumerate(osl_project_tree):
        if node["is_root"]:
            treeview_items.append(
                {
                    "id": node["uid"],
                    "text": node["name"],
                    "expanded": True,
                    "prefixIcon": {
                        "src": "https://s2.svgbox.net/materialui.svg?ic=account_tree"
                    },
                    "level": 0,
                    "children": []
                }
            )
        else:
            # Find parent node
            matching_indexes = find_dicts_by_key_recursively(treeview_items, (node["level"] - 1))
            if len(matching_indexes) == 0:
                raise Exception("Unable to find parent node.")
            parent_node = get_dict_from_indexes_sequence(treeview_items, matching_indexes[-1])
            if node["kind"] == "system":
                icon = "https://s2.svgbox.net/materialui.svg?ic=workspaces_filled"
            elif node["kind"] == "actor":
                icon = "https://s2.svgbox.net/materialui.svg?ic=workspaces_outline"
            else:
                raise ValueError(f"Unknown actor kind {node['kind']}.")
            parent_node["children"].append(
                {
                    "id": node["uid"],
                    "text": node["name"],
                    "expanded": True,
                    "prefixIcon": {
                        "src": icon
                    },
                    "level": node["level"],
                    "children": []
                },
            )

    return treeview_items


def check_empty_strings(lst) -> bool:
    for sublist in lst:
        for item in sublist:
            if not item.strip():  # Using strip() to remove leading/trailing whitespaces
                return False
    return True


def sort_dict_by_ordered_keys(dictionary: dict, list_of_keys: list) -> dict:
    return {key: dictionary[key] for key in list_of_keys if key in dictionary}


def sorted_nicely(l) -> Iterable:
    """Sort the given iterable in the way that humans expect.
    Reference: https://stackoverflow.com/questions/2669059/how-to-sort-alpha-numeric-set-in-python
    """

    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in re.split("([0-9]+)", key)]
    return sorted(l, key=alphanum_key)


def convert_microseconds(duration: int) -> str:
    """Convert a time duration in microseconds into a string with format: WWh XXmin YYs ZZms."""

    milliseconds = duration // 1000
    microseconds = duration % 1000
    seconds = milliseconds // 1000
    milliseconds = milliseconds % 1000
    minutes = seconds // 60
    seconds = seconds % 60
    hours = minutes // 60
    minutes = minutes % 60

    return f"{int(hours):02}h {int(minutes):02}m {int(seconds):02}s {int(milliseconds):03}ms"


def remove_key_from_dictionaries(dictionaries: list, key: str) -> list:

    for dictionary in dictionaries:
        if key in dictionary:
            del dictionary[key]

    return dictionaries


def update_list_of_tabs(node_info: dict) -> list:

    list_of_tabs = []

    for tab_info in MONITORING_TABS:
        if (
            node_info["is_root"]
            and tab_info["is_root"]
            or not node_info["is_root"]
            and node_info["kind"] == "system"
            and tab_info["is_system"]
            or not node_info["is_root"]
            and node_info["kind"] == "actor"
            and tab_info["is_actor"]
        ):
            list_of_tabs.append(
                dbc.Tab(
                    label=tab_info["label"],
                    tab_id=tab_info["tab_id"],
                    label_style={
                        "color": "#000000",
                        "text-color": "#000000",
                    },
                    active_label_style={
                        "color": "#FFFFFF",
                        "text-color": "#000000",
                        "background-color": "#000000",
                        "border-style": "solid",
                        "border-color": "#000000",
                    },
                )
            )

    return list_of_tabs


def extract_dict_by_key(dictionaries: list, key: str, value: Any, expect_unique: bool = False, return_index: bool = True) -> Union[None, dict]:
    """Given a list of dictionaries, return the ones matching specific key and value."""

    matches, indexes = [], []

    for (index, dictionary) in enumerate(dictionaries):
        if dictionary.get(key) == value:
            matches.append(dictionary)
            indexes.append(index)

    if expect_unique:
        if len(matches) == 0:
            raise Exception(f"No matching dictionaries with key {key} and value {value}.")
        elif len(matches) == 1:
            result_m = matches[0]
            result_i = indexes[0]
        else:
            raise Exception(f"Multiple matches found with key {key} and value {value}.")
    else:
        result_m = matches
        result_i = indexes

    return (result_m, result_i) if return_index else result_m


def read_log_file(log_file: Union[Path, str]) -> list:

    logs = []

    log_file = Path(log_file)

    if log_file.exists():
        with open(log_file, "r") as file:
            for line in file:
                if len(line.strip().rstrip("\n").replace(" ", "")) > 0:
                    logs.append(line.strip().rstrip("\n"))
    else:
        raise FileNotFoundError(f"Unable to find log file {log_file}.")

    return logs


def update_placeholders(ui_values: list, placeholders: dict) -> dict:
    """Information needed."""

    updated_dict = {}
    for row in ui_values[1:]:
        parameter_name = row["props"]["children"][0]["props"]["children"]["props"]["children"]
        input_value = row["props"]["children"][1]["props"]["children"]["props"]["value"]

        placeholder_values = placeholders.get("placeholder_values")
        if parameter_name in placeholder_values:
            updated_dict[parameter_name] = input_value
    return updated_dict


def check_optislang_server(osl_server: OslServer) -> None:
    """optiSLang server health check."""

    try:
        server_is_alive = osl_server.get_server_is_alive()
    except Exception as e:
        return False

    return server_is_alive


def get_states_ids_from_states(actor_states: dict) -> Tuple[str]:
    """Get available actor states ids from actor states response."""
    if not actor_states.get("states", None):
        return tuple([])
    return tuple([state["hid"] for state in actor_states["states"]])



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the monitoring step."""

import dash_bootstrap_components as dbc
import dash_daq as daq

from dash_extensions.enrich import Input, Output, State, html
from dash.exceptions import PreventUpdate

from ansys.saf.glow.client.dashclient import DashClient, callback

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.views import (
    design_table_view,
    project_summary_view,
    scenery_view,
    status_overview_view,
    summary_view,
    visualization_view,
)
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.utilities.common_functions import extract_dict_by_key, update_list_of_tabs


def layout(problem_setup_step: ProblemSetupStep, monitoring_step: MonitoringStep) -> html.Div:

    actor_info = extract_dict_by_key(problem_setup_step.osl_project_tree, "uid", monitoring_step.selected_actor_from_treeview, expect_unique=True, return_index=False)
    list_of_tabs = update_list_of_tabs(actor_info)

    return html.Div(
        [
            dbc.Stack(
                [
                    daq.BooleanSwitch(
                        id='activate_auto_update',
                        on=monitoring_step.auto_update_activated,
                        color="#FFB71B",
                        className="ms-auto",
                    ),
                    html.Div("Auto update")
                ],
                direction="horizontal",
                gap=1,
            ),
            dbc.Tabs(
                list_of_tabs,
                id="monitoring_tabs",
                active_tab=list_of_tabs[0].tab_id,
                style={
                    "color": "#000000",
                    "background-color": "#FFFFFF",
                    "text-color": "#000000",
                },
            ),
            html.Div(
                id="monitoring_page_content",
            )
        ]
    )


@callback(
    Output("monitoring_page_content", "children"),
    Input("monitoring_tabs", "active_tab"),
    Input("selected_state_dropdown", "value"),
    State("url", "pathname"),
)
def update_page_content(selected_tab, selected_state_id, pathname):
    """Update the page content according to the selected tab."""

    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step
    monitoring_step = project.steps.monitoring_step

    monitoring_step.selected_state_id = selected_state_id

    if selected_tab == "project_summary_tab":
        return project_summary_view.layout(problem_setup_step, monitoring_step)
    elif selected_tab == "summary_tab":
        return summary_view.layout(problem_setup_step, monitoring_step)
    elif selected_tab == "scenery_tab":
        return scenery_view.layout(problem_setup_step)
    elif selected_tab == "design_table_tab":
        return design_table_view.layout(problem_setup_step, monitoring_step)
    elif selected_tab == "visualization_tab":
        return visualization_view.layout(problem_setup_step, monitoring_step)
    elif selected_tab == "status_overview_tab":
        return status_overview_view.layout(problem_setup_step)


@callback(
    Output("activate_auto_update", "disabled"),
    Input("activate_auto_update", "on"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def activate_auto_update(on, pathname):
    """Enable/Disable auto update of monitoring tabs."""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    monitoring_step = project.steps.monitoring_step

    monitoring_step.auto_update_activated = on

    raise PreventUpdate



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the summary view."""

import dash_bootstrap_components as dbc
import json
import pandas as pd
from json.decoder import JSONDecodeError
import logging

from dash.exceptions import PreventUpdate
from dash_extensions.enrich import html, Input, Output, State, dcc
from ansys.saf.glow.client.dashclient import DashClient, callback
from ansys.saf.glow.core.method_status import MethodStatus

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.datamodel import datamodel
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.actor_information_table import ActorInformationTableAIO
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.button_group import ButtonGroup
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.node_control import NodeControlAIO
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.actor_logs_table import ActorLogsTableAIO
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.actor_statistics_table import ActorStatisticsTableAIO
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.utilities.common_functions import extract_dict_by_key, LOG_MESSAGE_COLORS


def layout(problem_setup_step: ProblemSetupStep, monitoring_step: MonitoringStep) -> html.Div:
    """Layout of the summary view."""
    # Get project data
    project_data = json.loads(problem_setup_step.project_data_file.read_text())
    # Get actor info
    actor_info = extract_dict_by_key(problem_setup_step.osl_project_tree, "uid", monitoring_step.selected_actor_from_treeview, expect_unique=True, return_index=False)
    # Get actor uid
    actor_uid = monitoring_step.selected_actor_from_treeview
    # Get actor hid
    actor_hid = monitoring_step.selected_state_id

    alert_props = {
        "children": monitoring_step.actor_command_execution_status["alert-message"],
        "color": monitoring_step.actor_command_execution_status["alert-color"],
        "is_open": bool(monitoring_step.actor_command_execution_status["alert-message"]),
    }
    button_group = ButtonGroup(options=monitoring_step.actor_btn_group_options, disabled=monitoring_step.commands_locked).buttons

    # Collect node-specific data
    if monitoring_step.selected_state_id:
        actor_information_data = project_data["actors"][actor_uid]["information"][actor_hid]
    else:
        actor_information_data = datamodel.extract_actor_information_data({}, {}, actor_info["kind"])
    actor_log_data = project_data["actors"][actor_uid]["log"]
    actor_statistics_data = project_data["actors"][actor_uid]["statistics"]

    # Assemble UI components
    content = [
        html.Br(),
        html.Div(
            id="alerts_container",
            style={"position": "fixed", "top": 130, "right": 10, "width": 350},
        ),
        dbc.Row(
            [
                dbc.Col(
                    html.Div(
                        ActorInformationTableAIO(
                            actor_information_data,
                        ),
                        id="actor_information_table"
                    ),
                    width=12
                ),
            ]
        ),
        dbc.Row(
            [
                dbc.Col(
                    NodeControlAIO(
                        button_group,
                        alert_props,
                        "actor-commands"
                    ),
                    width=12
                ),
            ]
        ),
        dbc.Row(
            [
                dbc.Col(
                    ActorLogsTableAIO(
                        actor_log_data,
                        monitoring_step.selected_page,
                        aio_id = "actor_logs_table"
                    ),
                    width=12
                ),
            ]
        ),
        dbc.Row(
            [
                dbc.Col(
                    html.Div(
                        ActorStatisticsTableAIO(
                            actor_statistics_data,
                        ),
                        id = "actor_statistics_table"
                    ),
                    width=12
                ),
            ]
        ),
        dcc.Interval(
            id="summary_auto_update",
            interval=monitoring_step.auto_update_frequency,  # in milliseconds
            n_intervals=0,
            disabled=False if monitoring_step.auto_update_activated else True
        ),
        html.Div(id="dummy-output", style= {'display': 'block'}),
    ]

    # Build layout
    return html.Div(content)


@callback(
    Output("alerts_container", "children"),
    Input("summary_auto_update", "n_intervals"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def display_alerts(n_intervals, pathname):
    # Get project
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    # Get problem setup step
    problem_setup_step = project.steps.problem_setup_step

    alerts_container = []

    method_state = problem_setup_step.get_long_running_method_state("start_and_monitor_osl_project")
    if method_state.status == MethodStatus.Failed:
        alerts_container.append(
            dbc.Toast(
                method_state.exception_message,
                header="Error",
                is_open=True,
                dismissable=True,
                icon="danger",
            )
        )

    if len(problem_setup_step.alerts) > 0:
        if problem_setup_step.alerts:
            alerts_container.append(
                dbc.Toast(
                    problem_setup_step.alerts["message"],
                    header=problem_setup_step.alerts["level"].capitalize(),
                    is_open=True,
                    dismissable=True,
                    icon=LOG_MESSAGE_COLORS[problem_setup_step.alerts["level"]],
                )
            )
        return alerts_container
    else:
        raise PreventUpdate


@callback(
    Output("actor_information_table", "children"),
    Output(ActorLogsTableAIO.ids.datatable("actor_logs_table"), "data"),
    Output(ActorLogsTableAIO.ids.datatable("actor_logs_table"), "store_value"),
    Output("actor_statistics_table", "children"),
    Output("selected_state_dropdown", "options"),
    Output("selected_state_dropdown", "value"),
    Output("summary_auto_update", "disabled"),
    Input("summary_auto_update", "n_intervals"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def update_view(n_intervals, pathname):
    """Update design table."""
    # Get project
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    # Get problem setup step
    problem_setup_step = project.steps.problem_setup_step
    # Get monitoring step
    monitoring_step = project.steps.monitoring_step

    if monitoring_step.auto_update_activated:
        try:
            # Get project data
            project_data = json.loads(problem_setup_step.project_data_file.read_text())
            # Get actor info
            actor_info = extract_dict_by_key(problem_setup_step.osl_project_tree, "uid", monitoring_step.selected_actor_from_treeview, expect_unique=True, return_index=False)
            # Get actor uid
            actor_uid = monitoring_step.selected_actor_from_treeview
            # Get actor hid
            actor_hid = monitoring_step.selected_state_id
            # Collect actor information data
            if monitoring_step.selected_state_id:
                actor_information_data = project_data.get("actors").get(actor_uid).get("information").get(actor_hid)
            else:
                actor_information_data = datamodel.extract_actor_information_data({}, actor_info["kind"])
            # Collect actor log data
            actor_log_data = project_data.get("actors").get(actor_uid).get("log")
            # Collect actor statistics data
            actor_statistics_data = project_data.get("actors").get(actor_uid).get("statistics")
            # Collect states ids
            if not monitoring_step.selected_state_id:
                if len(project_data.get("actors").get(monitoring_step.selected_actor_from_treeview).get("states_ids")):
                    monitoring_step.selected_state_id = project_data.get("actors").get(monitoring_step.selected_actor_from_treeview).get("states_ids")[0]
            return (
                ActorInformationTableAIO(actor_information_data),
                pd.DataFrame(actor_log_data).to_dict('records'),
                monitoring_step.selected_page,
                ActorStatisticsTableAIO(actor_statistics_data),
                project_data.get("actors").get(monitoring_step.selected_actor_from_treeview).get("states_ids"),
                monitoring_step.selected_state_id,
                True if problem_setup_step.osl_project_state in ["NOT STARTED", "FINISHED", "ABORTED"] else False
            )
        except Exception as e:
            logging.error(str(e))
            raise PreventUpdate
    else:
        raise PreventUpdate


@callback(
    Output("dummy-output", 'children'),
    Input(ActorLogsTableAIO.ids.datatable("actor_logs_table"), "page_current"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def update_table(page_current, pathname):

    # Get project
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)

    # Get monitoring step
    monitoring_step = project.steps.monitoring_step
    monitoring_step.selected_page = page_current
    return f"Dummy output updated {page_current}"


# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the problem setup step."""

import dash_bootstrap_components as dbc
import os
import time

from dash.exceptions import PreventUpdate
from dash_extensions.enrich import Input, Output, State, dcc, html, ALL, MATCH, no_update, callback_context
from ansys.saf.glow.client.dashclient import DashClient, callback
from ansys.saf.glow.core.method_status import MethodStatus
from ansys.solutions.optislang.frontend_components.load_sections import to_dash_sections, update_designs_to_dash_section
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.utilities.common_functions import check_empty_strings, LOG_MESSAGE_COLORS


def layout(problem_setup_step: ProblemSetupStep) -> html.Div:
    """Layout of the problem setup step."""
    if problem_setup_step.ui_placeholders and not problem_setup_step.project_locked:
        while problem_setup_step.get_long_running_method_state("update_osl_placeholders_with_ui_values").status == MethodStatus.Running: # current workaround to avoid raising ConflictError: {"detail":"update_osl_placeholders_with_ui_values is already running"} # current workaround to avoid raising ConflictError: {"detail":"update_osl_placeholders_with_ui_values is already running"}
            time.sleep(0.1)
        problem_setup_step.update_osl_placeholders_with_ui_values()
    project_properties_sections = to_dash_sections(
            problem_setup_step.placeholders, problem_setup_step.registered_files, problem_setup_step.project_locked
        )

    return html.Div(
        [
            # Header
            html.H1(
                problem_setup_step.app_metadata["title"],
                className="display-3",
                style={"font-size": "45px"},
            ),
            html.P(
                problem_setup_step.app_metadata["description"],
                className="lead",
                style={"font-size": "25px"},
            ),
            html.Hr(className="my-2"),
            html.Br(),
            # Alerts
            dbc.Row(
                id="project-locked-alert",
                children=[
                    dbc.Alert(
                        "The selected inputs are locked. To set new inputs, create a new project.",
                        color="info",
                    )
                ],
                style={"display": "block"} if problem_setup_step.project_locked else {"display": "none"},
            ),
            html.Div(
                id="alerts_container",
                style={"position": "fixed", "top": 130, "right": 10, "width": 350},
            ),
            html.Br(),
            # Input form
            html.P("Input Form", style={"font-size": "18px"}),
            html.Hr(className="my-2"),
            dbc.Row(id="osl-dash-sections",children=project_properties_sections),
            html.Br(),
            # Start analysis
            html.P("Start Analysis", style={"font-size": "18px"}),
            html.Hr(className="my-2"),
            html.Br(),
            dbc.Row(
                [
                    html.Div(
                        [
                            dbc.Button(
                                html.I(className="fas fa-play", style={"display": "inline-block"}),
                                id="start_analysis",
                                disabled=problem_setup_step.analysis_locked,
                                style = {
                                    "display": "flex",
                                    "justify-content": "center",
                                    "align-items": "center",
                                    "fontSize": "150%",
                                    "background-color": "rgba(0, 0, 0, 1)",
                                    "border-color": "rgba(0, 0, 0, 1)",
                                    "height": "30px",
                                }
                            ),
                            dbc.Tooltip(
                                "Start optiSLang project execution.",
                                target="start_analysis",
                            ),
                        ],
                        className="d-grid gap-2 col-5 mx-auto",
                    ),
                    dcc.Loading(
                        type="circle",
                        fullscreen=True,
                        color="#ffb71b",
                        style={
                            "background-color": "rgba(55, 58, 54, 0.1)",
                        },
                        children=html.Div(id="wait_start_analysis"),
                    ),
                    dcc.Interval(
                        id="problem_setup_alerts_update",
                        interval=1 * 3000,  # in milliseconds
                        n_intervals=0,
                    ),
                    dbc.Button(
                        id="dummy_button_2",
                        style={"display": "none"},
                        n_clicks=0,
                        color="dark"
                    ),
                ]
            ),
        ]
    )


@callback(
    Output("trigger_treeview_display", "data"),
    Output("navigation_tree", "items"),
    Output("wait_start_analysis", "children"),
    Output("start_analysis", "disabled"),
    Output("osl-dash-sections", "children"),
    Output("project-locked-alert", "style"),
    Output("alerts_container", "children"),
    Output("problem_setup_alerts_update", "disabled"),
    Input("start_analysis", "n_clicks"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def start_analysis(n_clicks, pathname):
    """Start optiSLang project."""
    if n_clicks:
        project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
        problem_setup_step = project.steps.problem_setup_step
        monitoring_step = project.steps.monitoring_step

        ui_data = problem_setup_step.ui_placeholders
        ui_data.update({"start_analysis_requested": True})
        problem_setup_step.ui_placeholders = ui_data

        trigger_treeview_display = False
        disable_start_analysis = False
        osl_dash_sections = []
        project_locked_alert = {"display": "none"}
        alerts_container = []
        disable_problem_setup_alerts_update = True

        # Check Ansys ecosystem
        try:
            problem_setup_step.check_ansys_ecosystem()
        except Exception as e:
            alerts_container.append(
                dbc.Toast(
                    str(e),
                    header="Error",
                    is_open=True,
                    dismissable=True,
                    icon="danger",
                ),
            )

        if problem_setup_step.get_method_state("check_ansys_ecosystem").status == MethodStatus.Completed:
            # Update project properties file prior to the solve
            problem_setup_step.write_updated_properties_file()
            # Start analysis
            problem_setup_step.start_and_monitor_osl_project()
            # Lock start analysis and ui data
            problem_setup_step.analysis_locked = True
            problem_setup_step.project_locked = True
            # Wait until the analysis effectively starts
            while True:
                method_state = problem_setup_step.get_long_running_method_state("start_and_monitor_osl_project")
                if method_state.status == MethodStatus.Running:
                    if problem_setup_step.osl_project_state != "NOT STARTED":
                        trigger_treeview_display = True
                        disable_start_analysis = True
                        osl_dash_sections = to_dash_sections(problem_setup_step.placeholders, problem_setup_step.registered_files, problem_setup_step.project_locked)
                        project_locked_alert = {"display": "none"}
                        alerts_container = []
                        disable_problem_setup_alerts_update = False
                        monitoring_step.auto_update_activated = True
                        break
                elif method_state.status == MethodStatus.Failed:
                    alerts_container.append(
                        dbc.Toast(
                            method_state.exception_message,
                            header="Error",
                            is_open=True,
                            dismissable=True,
                            icon="danger",
                        )
                    )
                    break

        return (
            trigger_treeview_display,
            problem_setup_step.treeview_items,
            True,
            disable_start_analysis,
            osl_dash_sections,
            project_locked_alert,
            alerts_container,
            disable_problem_setup_alerts_update
        )
    else:
        raise PreventUpdate


@callback(
    Output("alerts_container", "children"),
    Input("problem_setup_alerts_update", "n_intervals"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def display_alerts(n_intervals, pathname):
    # Get project
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    # Get problem setup step
    problem_setup_step = project.steps.problem_setup_step

    alerts_container = []

    method_state = problem_setup_step.get_long_running_method_state("start_and_monitor_osl_project")
    if method_state.status == MethodStatus.Failed:
        alerts_container.append(
            dbc.Toast(
                method_state.exception_message,
                header="Error",
                is_open=True,
                dismissable=True,
                icon="danger",
            )
        )

    if len(problem_setup_step.alerts) > 0:
        if problem_setup_step.alerts:
            alerts_container.append(
                dbc.Toast(
                    problem_setup_step.alerts["message"],
                    header=problem_setup_step.alerts["level"].capitalize(),
                    is_open=True,
                    dismissable=True,
                    icon=LOG_MESSAGE_COLORS[problem_setup_step.alerts["level"]],
                )
            )
        return alerts_container
    else:
        raise PreventUpdate


@callback(
    Output("start_analysis", "disabled"),
    Input("dummy_button_2", "n_clicks"),
    State({"container": "placeholders", "placeholder": ALL}, "value"),
    State({"container": "placeholders", "placeholder": ALL}, "id"),
    State({"type": "upload", "placeholder": ALL}, "id"),
    State("dummy_button_2", "n_clicks"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def initialize_dictionary_of_ui_placeholders(n_clicks, data, ids, input_file_ids, dummy_clicks, pathname):
    """This initializes the dictionary of ui placeholders using only the values in the Placeholders section."""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step
    designs = {}
    parameters = {}
    if problem_setup_step.ui_placeholders == {}:
        ui_data = problem_setup_step.ui_placeholders
        for index in range(len(data)):
            if ids[index]["placeholder"].startswith("StartDesigns"):
                split_values = ids[index]["placeholder"].split("##")
                values = split_values[1:]
                rowId = values[0].split("#")[1]
                key = values[1].split("#")[1]
                value = values[2].split("#")[1]
                if rowId in designs:
                    designs[rowId].update({key: value})
                else:
                    designs[rowId] = {key: value}
            elif ids[index]["placeholder"].startswith("ParameterManager"):
                split_values = ids[index]["placeholder"].split("##")
                values = split_values[1:]
                placeholder_name = values[0].split("#")[0]
                key = values[0].split("#")[1]
                if isinstance(data[index], list) and True in data[index]:
                    parameters[key] = True
                elif isinstance(data[index], list) and False in data[index]:
                    parameters[key] = False
                else:
                    parameters[key] = data[index]
                ui_data.update({placeholder_name: parameters})
            elif isinstance(data[index], list):
                value = data[index]
                if True in value:
                    new_value = True
                else:
                    new_value = False
                ui_data.update({ids[index]["placeholder"]: new_value})
            else:
                ui_data.update({ids[index]["placeholder"]: data[index]})
            ui_data.update({"StartDesigns": designs})
            ui_data["start_analysis_requested"] = False
        if input_file_ids:
            problem_setup_step.analysis_locked = True
        else:
            problem_setup_step.analysis_locked = False
        for index in range(len(input_file_ids)):
            ui_data.update({input_file_ids[index]["placeholder"]: ""})

        problem_setup_step.ui_placeholders = ui_data
        problem_setup_step.update_osl_placeholders_with_ui_values()
        return problem_setup_step.analysis_locked
    else:
        return no_update


@callback(
    Output({"container": "placeholders", "placeholder": MATCH}, "value"),
    Input({"container": "placeholders", "placeholder": MATCH}, "value"),
    State({"container": "placeholders", "placeholder": MATCH}, "id"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def update_ui_placeholders(value, id, pathname):
    """This updates the dictionary of ui placeholders each time the ui data changes in the Placeholders section. If the value entered is None, "", or invalid, the last value entered in the component is returned."""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step
    name = id["placeholder"]
    ui_data = problem_setup_step.ui_placeholders
    if value is None or value == "":
        if name.startswith("StartDesigns"):
            split_values = id["placeholder"].split("##")
            values = split_values[1:]
            rowId = values[0].split("#")[1]
            key = values[1].split("#")[1]
            last_value = ui_data["StartDesigns"][rowId][key]
        elif name.startswith("ParameterManager"):
            split_values = id["placeholder"].split("##")
            values=split_values[1:]
            pm_name = values[0].split("#")[0]
            key = values[0].split("#")[1]
            last_value = ui_data[pm_name][key]
        else:
            last_value = ui_data[name]
        return last_value
    else:

        if name.startswith("StartDesigns"):
            split_values = id["placeholder"].split("##")
            values = split_values[1:]
            rowId = values[0].split("#")[1]
            key = values[1].split("#")[1]
            ui_data["StartDesigns"][rowId][key] = value
        elif name.startswith("ParameterManager"):
            split_values = id["placeholder"].split("##")
            values=split_values[1:]
            pm_name = values[0].split("#")[0]
            key = values[0].split("#")[1]
            if isinstance(value, list) and True in value:
                value = True
            elif isinstance(value, list) and False in value:
                value = False
            ui_data[pm_name][key] = value
        elif isinstance(value, list):
            if True in value:
                value = True
            else:
                value = False
            ui_data.update({name: value})
        else:
            ui_data.update({name: value})
        problem_setup_step.ui_placeholders = ui_data
        return no_update


@callback(
    Output({"type": "upload", "placeholder": MATCH}, "children"),
    Input({"type": "upload", "placeholder": MATCH}, "isCompleted"),
    State({"type": "upload", "placeholder": MATCH}, "fileNames"),
    State({"type": "upload", "placeholder": MATCH}, "upload_id"),
    State({"type": "upload", "placeholder": MATCH}, "id"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def upload_input_files_to_project_and_update_ui_placeholders(is_completed, filenames, upload_id, component_id, pathname):
    """This uploads an Input file to the project directory and updates the dictionary of ui placeholders every time
    the ui data changes in the Input files section."""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step

    name = component_id["placeholder"]
    if is_completed and filenames and len(filenames):
        filename = filenames[0]
        filepath = os.path.join(problem_setup_step.upload_directory, upload_id, filename)
        file_data = open(filepath, mode="rb")

        project.upload_file(f"Problem_Setup/Input_Files/{filename}", file_data)
        ui_data = problem_setup_step.ui_placeholders
        ui_data.update({name: filename})
        problem_setup_step.ui_placeholders = ui_data

        return no_update
    else:
        raise PreventUpdate


@callback(
    Output("start_analysis", "disabled"),
    Input({"type": "upload", "placeholder": ALL}, "isCompleted"),
    State({"type": "upload", "placeholder": ALL}, "fileNames"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def update_status_of_start_analysis_button(is_completed, filenames, pathname):
    """This enables the start analysis button if a file has been uploaded for all input files."""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step
    if all(filenames):
        problem_setup_step.analysis_locked = False
    else:
        problem_setup_step.analysis_locked = True
    if filenames == True:
        result = check_empty_strings(filenames)

    return problem_setup_step.analysis_locked


@callback(
    Output("dummy_button_2", "n_clicks"),
    Input("url", "pathname"),
    State("dummy_button_2", "n_clicks"),
)
def on_page_load_initialize_dictionary_of_ui_placeholder_values(pathname, n_clicks):
    return n_clicks + 1


@callback(
    Output("start-designs-view", "children"),
    Input({"type": "button-add", "index": ALL}, "n_clicks"),
    Input({"type": "button-del", "index": ALL}, "n_clicks"),
    State({"type": "button-del", "index": ALL}, "disabled"),
    State({"type": "row", "index": ALL}, "id"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def update_start_designs_table(n_clicks_add, n_clicks_del, disabled_states, row_id, pathname):
    """This updates the StartDesigns table every time the '+' or '-' button is clicked."""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step
    ctx = callback_context
    triggered_button = ctx.triggered[0]["prop_id"].split(".")[0]
    if (
        ctx.triggered
        and not all(n_click is None for n_click in n_clicks_del)
        or not all(n_click is None for n_click in n_clicks_add)
    ):
        ui_data = problem_setup_step.ui_placeholders
        designs = ui_data["StartDesigns"]
        button_type = triggered_button.split('"type":"')[1].split('"}')[0]
        row_id = triggered_button.split('"index":')[1].split(",")[0]
        if button_type == "button-add":
            designs = add_row(designs, row_id)
        elif button_type == "button-del":
            designs = delete_row(designs, row_id)

        # Rebuild the dictionary with updated keys
        new_designs = {str(index): value for index, (_, value) in enumerate(designs.items(), start=1)}
        ui_data.update({"StartDesigns": new_designs})

        problem_setup_step.ui_placeholders = ui_data
        while problem_setup_step.get_long_running_method_state("update_osl_placeholders_with_ui_values").status == MethodStatus.Running: # current workaround to avoid raising ConflictError: {"detail":"update_osl_placeholders_with_ui_values is already running"}
            time.sleep(0.1)
        problem_setup_step.update_osl_placeholders_with_ui_values()

        return update_designs_to_dash_section(problem_setup_step.placeholders, problem_setup_step.project_locked)
    else:
        raise PreventUpdate


def add_row(designs, id_):
    new_value = designs[id_]  # Value from the target key
    new_designs = {}
    for key, value in designs.items():
        new_designs[key] = value
        if key == id_:
            new_designs["temp"] = new_value
    return new_designs


def delete_row(designs, id_):
    del designs[id_]
    return designs



{%- if cookiecutter.copyright != "None" -%}
# Copyright (c) {% now "utc", '%Y' %}, {{ cookiecutter.copyright }}. Unauthorised use, distribution or duplication is prohibited
{% endif %}

"""
{{ cookiecutter.project_name }}.

{{ cookiecutter.library_name }}
"""

import os
from pathlib import Path
import sys

sys.path.insert(0, os.path.join(Path(os.path.dirname(os.path.realpath(__file__))).parent, "stubs"))
